/* :ts=4
 *
 * Amiga SnipIt 1.2
 * (c) (opyright 1987,1988 - Scott Evernden - All Rights Reserved
 *
 * main.c - cutter/paster like SunTools
 *
 * Permission is granted to modify and redistribute as long
 * as this notice is left intact.
 *
 */

/**********************

	The strategy here begins with the installation of some input handler
	code into the input event handler chain.  This code looks for mouse
	actions and keyboard presses that indicate that text is being snipped,
	or is being pasted.

	Snipped text (plane 0) is always immediately copied into a temp 1-plane
	bitmap without further processing.  Subsequents "snips" cause the old
	snipped bitmap to be thrown away, and a new one captured.

	Pasting requires that the text in the capture bitmap be recognized.
	This is not done in the input handler code, since it would cause all
	pending input events to pile up while the recognizer is running.
	Instead, the input handler code signals the main snipit code (RUNning
	in the background, itself) to do the recognition and fake the kb events.
	During this period, the input handler code temporarily suspends its
	special processing waiting for the main code (recognizer) to finish.

	Once a capture bitmap has been recognized and converted into a sequence
	of key codes, it is deallocated, and subsequent pastes don't require
	the recognition step.

**********************/

#include "hs.h"

struct Task *_FindTask();
struct Library *_OpenLibrary();

BPTR output = 0;

long LayersBase;					/* for WhichLayer() */
struct GfxBase *GfxBase;
struct IntuitionBase *IntuitionBase;
struct ConsoleBase *ConsoleBase;	/* defined in hs.h */

/* communication with input.device */
struct IOStdReq *inpReq;
struct MsgPort *inpPort;

/* communication with console.device */
struct MsgPort *conPort;
struct IOStdReq *conReq;

/* who I am */
struct Task *myTask;

/* snipit's port */
struct MsgPort *myPort;

struct MyMessage {
	struct Message m;
	int done, huh, off;
	int w_offx, w_offy;
	int c_offx, c_offy;
	long alen;
	char *aptr;
} myMsg, *msgp;

/* for snipit's input handler */
struct Interrupt handler;

/* flags handler installation for cleanup */
int handlerIn;

/* snipit's allocated signal bit */
int mySignal = -1L;

/* snipit's input handler code */
VOID handlerCode();

/* snipit replacement functions */
VOID myRectFill(), myScrollRaster(), myCloseWindow();

/* remember original library vectors */
long oldRectFill, oldScrollRaster, oldCloseWindow;

/* vector offsets of certain functions */
extern int lvoRectFill, lvoScrollRaster, lvoCloseWindow;

/* detach stuff */
long running = 0;
char *_procname = "\x9B33mSnipIt 1.2\x9B0m";

char *huh_msg =   "Huh?  (for help type: snipit h)";

/* "Can't open"... forms */
char *i_libname = "~intuition.library";
char *g_libname = "~graphics.library";
char *l_libname = "~layers.library";
char *i_devname = "~input.device";
char *c_devname = "~console.device";

char *port_name = "^\x9B33mSnipIt 1.2\x9B0m Port";

long _savsp;

extern int jobFlag;
extern int con_offx, con_offy;
extern int win_offx, win_offy;
extern int cmd1, cmd2;
extern int snip, lock_layer, windows;
extern char preString[];

/*********************************************/
/* pick up the ABS symbols describing gfx/intui lib offsets
 */
#asm
	dseg
	public	_LVORectFill
	public	_LVOScrollRaster
	public	_LVOCloseWindow

_lvoRectFill		dc.w	_LVORectFill
_lvoScrollRaster	dc.w	_LVOScrollRaster
_lvoCloseWindow		dc.w	_LVOCloseWindow

	cseg
#endasm

/*********************************************/

writes(str)
register char *str;
{
	register char *piece;
	register long len;

	len = 0;

	if (*str == '!') {
		piece = "Can't ";
		len = 6;
	}
	else if (*str == '~') {
		piece = "Can't open ";
		len = 11;
	}
	else if (*str == '^') {
		piece = "Can't create ";
		len = 13;
	}
	else if (*str == '_') {
		piece = _procname;
		len = 17;
	}

	if (len) {
		str++;
		_Write(output, piece, len);
	}

	/* output a message & newline */
	_Write(output, str, (long) strlen(str));
	_Write(output, "\n", 1L);
}

/*********************************************/

fini()
{
	long stat;

	/* cleanup in other modules */
	finiHand();
	finiReco();

	/* restore gfx/intui library vectors */
	if (oldRectFill)
		SetFunction(GfxBase, (long) lvoRectFill, oldRectFill);

	if (oldScrollRaster)
		SetFunction(GfxBase, (long) lvoScrollRaster, oldScrollRaster);

	if (oldCloseWindow)
		SetFunction(IntuitionBase, (long) lvoCloseWindow, oldCloseWindow);

	/* remove snipit input handler */
	if (handlerIn) {
		inpReq->io_Command = IND_REMHANDLER;
		inpReq->io_Data = (APTR) &handler;
		DoIO(inpReq);
	}

	/* close down console.device */
	if (conReq) {
		if (conReq->io_Device)
			CloseDevice(conReq);
		DeleteStdIO(conReq);
	}
	if (conPort)
		DeletePort(conPort);

	/* close down input.device */
	if (inpReq) {
		if (inpReq->io_Device)
			CloseDevice(inpReq);
		DeleteStdIO(inpReq);
	}
	if (inpPort)
		DeletePort(inpPort);

	/* release snipit's port & signal bit */
	if (myPort)
		DeletePort(myPort);

	if (mySignal != -1)
		FreeSignal(1L << mySignal);

	/* close libraries */
	if (GfxBase)
		_CloseLibrary(GfxBase);

	if (LayersBase)
		_CloseLibrary(LayersBase);

	if (IntuitionBase)
		_CloseLibrary(IntuitionBase);

	if (output)
		_Close(output);

#asm
	moveq	#0,d0			;pick up return exit code
	move.l	__savsp,sp		;get back original stack pointer
	rts						;and exit
#endasm
}

/*********************************************/

int init()
{
	long stat;

	/* global record of this task's ID */
	myTask = _FindTask((char *) NULL);

	/* need intuitionbase to read mouse pos. & screen layer */
	IntuitionBase = (struct IntuitionBase *) _OpenLibrary(i_libname + 1, 0L);
	if (IntuitionBase == NULL) {
		writes(i_libname);
		return 0;
	}

	/* need graphics since we do */
	GfxBase = (struct GfxBase *) _OpenLibrary(g_libname + 1, 0L);
	if (GfxBase == NULL) {
		writes(g_libname);
		return 0;
	}

	/* layers for WhichLayer() */
	LayersBase = (long) _OpenLibrary(l_libname + 1, 0L);
	if (LayersBase == NULL) {
		writes(l_libname);
		return 0;
	}

	/* signal from input event handler to main code */
	mySignal = AllocSignal(-1L);
	if (mySignal == -1) {
		writes("!allocate a signal bit.");
		return 0;
	}

	/* a signal to this port will terminate us */
	myPort = CreatePort(port_name + 1, 0L);
	if (myPort == NULL) {
		writes(port_name);
		return 0;
	}

	/* port to the input.device */
	inpPort = CreatePort((char *) NULL, 0L);
	if (inpPort == NULL) {
		writes("^input.device port.");
		return 0;
	}

	/* need an i/o request */
	inpReq = CreateStdIO(inpPort);
	if (inpReq == NULL) {
		writes("^input.device i/o request.");
		return 0;
	}

	/* open input.device */
	stat = OpenDevice(i_devname + 1, 0L, inpReq, 0L);
	if (stat != 0) {
		writes(i_devname);
		return 0;
	}

	/* port to the console.device */
	conPort = CreatePort((char *) NULL, 0L);
	if (!conPort) {
		writes("^console.device port.");
		return 0;
	}

	/* need an i/o request */
	conReq = CreateStdIO(conPort);
	if (!conReq) {
		writes("^console.device i/o request.");
		return 0;
	}

	/* open console.device */
	stat = OpenDevice(c_devname + 1, -1L, conReq, 0L);
	if (stat != 0) {
		writes(c_devname);
		return 0;
	}

	/* check for consolebase */
	ConsoleBase = (struct ConsoleBase *) conReq->io_Device;
	if (!ConsoleBase) {
		writes("!find ConsoleBase.");
		return 0;
	}

	/* handler command stuff */
	handler.is_Code = handlerCode;
	handler.is_Node.ln_Pri = 51;

	/* install new input event handler */
	inpReq->io_Command = IND_ADDHANDLER;
	inpReq->io_Data = (APTR) &handler;

	stat = DoIO(inpReq);
	if (stat != 0) {
		writes("ADDHANDLER request failed.");
		return 0;
	}
	handlerIn = 1;

	/* function catchers */

	oldRectFill = SetFunction(GfxBase, (long) lvoRectFill, myRectFill);
	if (oldRectFill == 0) {
		writes("!replace RectFill()");
		return 0;
	}

	oldScrollRaster = SetFunction(GfxBase, (long) lvoScrollRaster,
									myScrollRaster);
	if (oldScrollRaster == 0) {
		writes("!replace ScrollRaster()");
		return 0;
	}

	oldCloseWindow = SetFunction(IntuitionBase, (long) lvoCloseWindow,
									myCloseWindow);
	if (oldCloseWindow == 0) {
		writes("!replace CloseWindow()");
		return 0;
	}

	return 1;
}

/*********************************************/

int jump(mode)
int mode;
{
	char ch;

	while (0 < msgp->alen--) {
		ch = *msgp->aptr++;
		if ((ch != ' ' && ch != '\t' && ch != '\n') == mode)
			return ch;
	}
	return 0;
}

/*********************************************/

int innum(base)
int base;
{
	int result, neg, ch;

	result = 0;
	ch = jump(1);
	if (!ch)
		return 0;

	neg = ch == '-';
	if ((neg || ch == '+') && 0 < msgp->alen--)
		ch = *msgp->aptr++;

	while (1) {
		ch -= '0';
		if (9 < ch)
			ch -= 7;	/* was 'A'-'F' */
		if (15 < ch)
			ch -= 32;	/* was 'a'-'f' */

		if (ch < 0 || base <= ch) {
			msgp->alen++;	/* ungetc */
			msgp->aptr--;
			break;
		}

		result = result * base + ch;

		if (msgp->alen-- <= 0)
			break;

		ch = *msgp->aptr++;
	}
	return neg ? -result : result;
}

/*********************************************/

usage()
{
writes("Usage: snipit [opts]");
writes("where opts can be any of:");
writes("   ? or h         Help; this text");
writes("   q              Quit; remove SnipIt from the system");
writes("   p \"prestring\"  When pasting using Command-B key, insert");
writes("                  \"prestring\" before each line.  In quotes");
writes("   l 1 or 0       Lock layer (or don't) while snipping");
writes("   u +/-number    Adjust cell x position in console windows");
writes("   v +/-number    Adjust cell y position in console windows");
writes("   x +/-number    Adjust cell x position in non-console windows");
writes("   y +/-number    Adjust cell y position in non-console windows");
writes("   a hex-number   Qualifier for Command-A key (default LEFT-AMIGA)");
writes("   b hex-number   Qualifier for Command-B key (default LEFT-ALT)");
writes("");
writes("Defaults are:  p\"> \" l0 u0 v0 x0 y0 a0040 b0010");
}

/*********************************************/

int options()
{
	char *cp;
	int temp, ch;

	msgp->done = msgp->huh = msgp->off = 0;

	do {
		/* locate a char after whitespace */
		ch = jump(1);					/* advance to 1st non-white */

		if ('A' <= ch && ch <= 'Z')
			ch = ch - 'A' + 'a';		/* tolower */

		/* act on char */
		switch (ch) {

		case 0:		break;

		case 'q':	msgp->done = 1;
					return 0;			/* die */

		case 'p':	cp = preString;
					ch = jump(1);		/* delim char */
					if (ch) {
						while (0 < msgp->alen-- &&
							   (*cp = *msgp->aptr++) != ch) {
							cp++;
							if (preString + PRE_SIZE <= cp)
								break;
						}
					}
					*cp = 0;
					break;

		case 'x':	if (snip == WIN_SNIP)
						clearHi();
					msgp->off = 1;
					win_offx += innum(10);
					break;

		case 'y':	if (snip == WIN_SNIP)
						clearHi();
					msgp->off = 1;
					win_offy += innum(10);
					break;

		case 'u':	if (snip == CON_SNIP)
						clearHi();
					msgp->off = 1;
					con_offx += innum(10);
					break;

		case 'v':	if (snip == CON_SNIP)
						clearHi();
					msgp->off = 1;
					con_offy += innum(10);
					break;

		case 'a':	if (temp = innum(16));
						cmd1 = temp;
					break;

		case 'b':	if (temp = innum(16));
						cmd2 = temp;
					break;

		case 'l':	ch = jump(1);		/* first char */
					if (ch)
						lock_layer = ch == '1';
					break;

		case 'w':	ch = jump(1);		/* first char */
					if (ch)
						windows = ch == '1';
					break;

		default:	msgp->huh = 1;		/* newline, too - end of string */
					break;
		}

	} while (jump(0));					/* advance to whitespace */

	if (msgp->off) {
		win_offx = win_offx < -32 ? -32 : 32 < win_offx ? 32 : win_offx;
		win_offy = win_offy < -32 ? -32 : 32 < win_offy ? 32 : win_offy;
		con_offx = con_offx < -32 ? -32 : 32 < con_offx ? 32 : con_offx;
		con_offy = con_offy < -32 ? -32 : 32 < con_offy ? 32 : con_offy;

		msgp->w_offx = win_offx;
		msgp->w_offy = win_offy;
		msgp->c_offx = con_offx;
		msgp->c_offy = con_offy;
	}

	return 1;	/* i.e., keep running */
}

/*********************************************/

outnum(n)
int n;
{
	char d;

	if (n < 0) {
		n = -n;
		_Write(output, "-", 1L);
	}
	if (n > 9) {
		d = n / 10 + '0';
		_Write(output, &d, 1L);
	}
	d = n % 10 + '0';
	_Write(output, &d, 1L);
}

/*********************************************/

currAdjust(type, x, y)
char *type;
int x, y;
{
	_Write(output, type, 7L);
	_Write(output, " adjustment now: x=", 19L);
	outnum(x);
	_Write(output, ",y=", 3L);
	outnum(y);
	_Write(output, "\n", 1L);
}

/*********************************************/

checkStatus()
{
	if (myMsg.huh)
		writes(huh_msg);

	else if (myMsg.done)
		writes("_ cancelled.");

	else if (myMsg.off) {
		currAdjust("Console", myMsg.c_offx, myMsg.c_offy);
		currAdjust("Window ", myMsg.w_offx, myMsg.w_offy);
	}
}

/*********************************************/

_main(alen, aptr)	/* DOSBase and ExecBase opened already by crt0.o */
long alen;
char *aptr;
{
	static int pass;
	int keep_going;
	long sige, sigt;
	struct MsgPort *port, *rport;

	/* sever seglists */
	do_detach(&alen, &aptr);

	myMsg.alen = alen;
	myMsg.aptr = aptr;

	/* help requested?? */
	if (*aptr == '?' || *aptr == 'h')
		usage();

	/* installed already? if so, pass arguments */
	else if (port = FindPort(port_name + 1)) {	

		if (alen <= 1)
			writes("_ already installed!");

		else {
			myMsg.m.mn_ReplyPort = rport = CreatePort(NULL, 0L);
			if (!rport)
				writes("^a port");

			else {
				PutMsg(port, &myMsg);
				_WaitPort(rport);		/* wait for reply */
				_GetMsg(rport);
				DeletePort(rport);

				/* check return stats */
				checkStatus();
			}
		}
	}

	/* install & deal with args */
	else if (init()) {

		msgp = &myMsg;
		if (options() && !myMsg.huh) {

			writes("_ installed.  \xA9 1988 - Scott Evernden");

			running = 1;

			/* signals which mean something */
			sige = 1L << mySignal;			/* request for recognition */
			sigt = 1L << myPort->mp_SigBit;	/* options message */

			/* process requests for handler code
			 * while awaiting options signal
			 */

			keep_going = 1;
			while (keep_going) {

				if (Wait(sige | sigt) == sige) {
					buildEvents();
					jobFlag = 0;		/* lower active flag */
				}
				else while (msgp = (struct MyMessage *) GetMsg(myPort)) {
					keep_going = options();
					_ReplyMsg(msgp);
				}
			}
		}

		else { 
			if (myMsg.huh)
				writes(huh_msg);

			writes("_ not installed.");
		}
	}

	running = 1;

	/* terminate */
	Delay(10L);
	fini();
}

